#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import base64
from datetime import datetime

import bcrypt
from ckanext.query_dois.model import QueryDOIStat
import uuid

# action types
DOWNLOAD_ACTION = 'download'
SAVE_ACTION = 'save'


def anonymize_email(email_address):
    """
    Split the email address into it's identity and domain parts, then return the secure
    hash of the identity and the domain. Bcrypt is used to securely hash the identity,
    using the domain as the salt.

    :param email_address: the email address
    :return: a 2-tuple of the email address and the domain
    """
    if email_address is None:
        return None, None

    email_address = email_address.lower()
    # figure out the domain from the email address
    try:
        domain = email_address[email_address.index('@') + 1 :]
    except ValueError:
        # no @ found, just use the whole string
        domain = email_address

    # create a custom salt by base64 encoding the domain and then trimming the whole thing to 22
    # characters (which is bcrypt's required salt length). Note that we fill the right side of the
    # domain with dots to ensure it's at least 18 characters in length. This is necessary as we need
    # to ensure that the base64 encode result is at least 22 characters long and 18 is the minimum
    # input length necessary to create a base64 encoding result of at least 22 characters.
    salt = b'$2b$12$' + base64.b64encode(domain.zfill(18).encode('utf-8'))[:22]
    return bcrypt.hashpw(email_address.encode('utf-8'), salt), domain


def record_stat(query_doi, action, email_address=None, domain=None, identifier=None):
    """
    Creates a new QueryDOIStat object and saves it to the database.

    :param query_doi: the QueryDOI object against which the stat should be stored
    :param action: the action that occurred to trigger this stat (for example: "download")
    :param email_address: the email address of the user performing the action
    :param domain: an alternate domain name if email not specified
    :param identifier: an alternate identifier if email not specified
    :return: a new QueryDOIStat object
    """
    if email_address:
        identifier, domain = anonymize_email(email_address)
    if identifier is None:
        # just a random uuid if nothing else is specified, so we don't end up grouping
        # many unrelated users together under the identifier of "None"
        identifier = uuid.uuid4().hex
    stat = QueryDOIStat(
        doi=query_doi.doi,
        action=action,
        domain=domain,
        identifier=identifier,
        timestamp=datetime.now(),
    )
    stat.save()
    return stat



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import itertools
from datetime import datetime
from functools import partial

from ckan import model
from ckan.plugins import toolkit
from sqlalchemy import false


def get_public_datastore_resources(only=None):
    """
    Retrieve all the public resource ids from the database that are also in the
    datastore. If the only parameter is provided, it is used to filter the return so
    that it only includes those in the only list.

    :param only: a list/set/whatever of resource ids to include in the returned set
    :return: a set of public resource ids
    """
    # retrieve all resource ids that are active, in an active package and in a public package
    query = (
        model.Session.query(model.Resource)
        .join(model.Package)
        .filter(model.Resource.state == 'active')
        .filter(model.Package.state == 'active')
        .filter(model.Package.private == false())
        .with_entities(model.Resource.id)
    )
    if only:
        query = query.filter(model.Resource.id.in_(list(only)))

    public_resource_ids = set()

    # cache this action (with context) so that we don't have to retrieve it over and over again
    is_datastore_resource = partial(
        toolkit.get_action('datastore_is_datastore_resource'), {}
    )
    for row in query:
        if is_datastore_resource(dict(resource_id=row.id)):
            public_resource_ids.add(row.id)

    return public_resource_ids


def get_invalid_resources(resource_ids):
    '''

    :param resource_ids:
    :return:
    '''
    resource_ids = set(resource_ids)
    public_resource_ids = get_public_datastore_resources(only=resource_ids)
    return resource_ids - public_resource_ids


def extract_resource_ids_and_versions(
    req_version=None, req_resource_ids=None, req_resource_ids_and_versions=None
):
    if req_resource_ids_and_versions is not None:
        req_resource_ids = set(req_resource_ids_and_versions.keys())
    else:
        req_resource_ids = (
            set(req_resource_ids) if req_resource_ids is not None else set()
        )

    resource_ids = get_public_datastore_resources(only=req_resource_ids)
    bad_resources = req_resource_ids - resource_ids
    if bad_resources:
        # resources were requested, but not all of them were public/active
        raise toolkit.ValidationError(
            f'Some of the resources requested are private or not active, '
            f'DOIs can only be created using public, active resources. '
            f'Invalid resources: {", ".format(bad_resources)}'
        )
    elif len(resource_ids) == 0:
        # no resources available
        raise toolkit.ValidationError('No public resources are available')

    version = req_version if req_version is not None else to_timestamp(datetime.now())
    # round all the versions down for each resource
    if req_resource_ids_and_versions is not None:
        iterator = req_resource_ids_and_versions.items()
    else:
        iterator = zip(resource_ids, itertools.repeat(version))

    round_version_action = partial(
        toolkit.get_action('datastore_get_rounded_version'), {}
    )
    resource_ids_and_versions = {}
    for resource_id, resource_version in iterator:
        data_dict = {'resource_id': resource_id, 'version': resource_version}
        rounded_version = round_version_action(data_dict)
        # this isn't really something that should happen, but if it does it just means there's no
        # data in the resource's datastore index, leave it out of the return dict
        if rounded_version is not None:
            resource_ids_and_versions[resource_id] = rounded_version
    return resource_ids_and_versions


def to_timestamp(moment: datetime) -> int:
    """
    Converts the given moment to a UNIX epoch in milliseconds.

    :param moment: a datetime object
    :return: integer UNIX epoch in milliseconds
    """
    ts = moment.timestamp()
    # multiply by 1000 to get the time in milliseconds and use int to remove any decimal places
    return int(ts * 1000)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import itertools
from functools import partial

from ckan import model
from ckan.plugins import toolkit
from sqlalchemy import false


def get_resource_and_package(resource_id):
    """
    Given a resource ID, returns the resource's dict and the parent package's dict too.

    :param resource_id: the resource ID
    :return: a 2-tuple, containing the resource dict and the package dict
    """
    resource = toolkit.get_action('resource_show')({}, {'id': resource_id})
    package = toolkit.get_action('package_show')({}, {'id': resource['package_id']})
    return resource, package


def get_public_datastore_resources(only=None):
    """
    Retrieve all the public resource ids from the database that are also in the
    datastore. If the only parameter is provided, it is used to filter the return so
    that it only includes those in the only list.

    :param only: a list/set/whatever of resource ids to include in the returned set
    :return: a set of public resource ids
    """
    # retrieve all resource ids that are active, in an active package and in a public package
    query = (
        model.Session.query(model.Resource)
        .join(model.Package)
        .filter(model.Resource.state == 'active')
        .filter(model.Package.state == 'active')
        .filter(model.Package.private == false())
        .with_entities(model.Resource.id)
    )
    if only:
        query = query.filter(model.Resource.id.in_(list(only)))

    public_resource_ids = set()

    # cache this action (with context) so that we don't have to retrieve it over and over again
    is_datastore_resource = partial(
        toolkit.get_action('datastore_is_datastore_resource'), {}
    )
    for resource_id in query:
        if is_datastore_resource(dict(resource_id=resource_id)):
            public_resource_ids.add(resource_id)

    return public_resource_ids


def get_authors(resource_ids):
    """
    Given some resource ids, return a list of unique authors from the packages
    associated with them.

    :param resource_ids: the resource ids
    :return: a list of authors
    """
    query = (
        model.Session.query(model.Resource)
        .join(model.Package)
        .filter(model.Resource.id.in_(list(resource_ids)))
        .with_entities(model.Package.author)
    )
    return list(set(itertools.chain.from_iterable(query)))


def get_resource_counts(query, query_version, resource_ids_and_versions):
    """
    Given a set of query parameters, figure out how many records are found on each
    resource.

    :param query: the query dict
    :param query_version: the query version
    :param resource_ids_and_versions: the resource ids and their specific versions
    :return: a dict of resource ids to counts
    """
    counts = {}
    multisearch_action = partial(toolkit.get_action('datastore_multisearch'), {})
    for resource_id, version in resource_ids_and_versions.items():
        # find out how many records match the query in the specific resource
        search_data_dict = {
            'query': query,
            'query_version': query_version,
            'resource_ids': [resource_id],
            'version': version,
            'size': 0,
        }
        counts[resource_id] = multisearch_action(search_data_dict)['total']
    return counts



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit


@toolkit.auth_allow_anonymous_access
def create_doi(context, data_dict):
    return {'success': True}



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK
import click
from ckan import model

from .model import query_doi_table, query_doi_stat_table


def get_commands():
    return [query_dois]


@click.group()
def query_dois():
    """
    Query DOIs CLI.
    """
    pass


@query_dois.command(name='initdb')
def init_db():
    """
    Creates the `query_doi` and `query_doi_stat` tables used by this extension.
    """
    # create the 2 tables if they don't already exist
    for table in (query_doi_table, query_doi_stat_table):
        if not table.exists(model.meta.engine):
            table.create(model.meta.engine)
            click.secho('Created "{}" table'.format(table), fg='green')
        else:
            click.secho(
                'Table "{}" already exists, skipping...'.format(table), fg='green'
            )



from ckanext.query_dois.plugin import QueryDOIsPlugin
from unittest.mock import MagicMock, patch


class TestIntegrationWithIVersionedDatastoreDownloads:
    @patch('ckanext.query_dois.plugin.record_stat')
    def test_download_email_context_is_modified(self, record_stat_mock):
        plugin = QueryDOIsPlugin()

        request = MagicMock()
        context = {}
        doi = MagicMock(doi='some/doi')

        find_existing_doi_mock = MagicMock(return_value=doi)

        with patch(
            'ckanext.query_dois.plugin.find_existing_doi', find_existing_doi_mock
        ):
            ret_context = plugin.download_modify_notifier_template_context(
                request, context
            )

        assert ret_context is context
        assert context['doi'] == doi.doi

    @patch('ckanext.query_dois.plugin.record_stat')
    def test_download_email_context_is_always_returned_when_find_errors(
        self, record_stat_mock
    ):
        plugin = QueryDOIsPlugin()

        request = MagicMock()
        context = {}

        find_existing_doi_mock = MagicMock(side_effect=Exception)

        with patch(
            'ckanext.query_dois.plugin.find_existing_doi', find_existing_doi_mock
        ):
            ret_context = plugin.download_modify_notifier_template_context(
                request, context
            )

        assert ret_context is context
        assert 'doi' not in context

    def test_download_email_context_contains_doi_if_we_get_one_even_if_error(self):
        """
        If the DOI gets generated we should stick it in the context as soon as possible.

        This means that even if less important calls fail (like the record_stat call)
        we'll get a doi back in the context. This test checks that functionality.
        """
        plugin = QueryDOIsPlugin()

        request = MagicMock()
        context = {}
        doi = MagicMock(doi='some/doi')

        find_existing_doi_mock = MagicMock(return_value=doi)
        record_stat_mock = MagicMock(side_effect=Exception)

        with patch('ckanext.query_dois.plugin.record_stat', record_stat_mock):
            with patch(
                'ckanext.query_dois.plugin.find_existing_doi', find_existing_doi_mock
            ):
                ret_context = plugin.download_modify_notifier_template_context(
                    request, context
                )

        assert ret_context is context
        assert context['doi'] == doi.doi
